﻿-----------------------------------------------------------------------------
--
-- data
--   \deaths
--     \<deathName>
--       - count
--
-- For non-grouped units:
--       \deathTimes (Array)
--
-- For grouped units:
--       \deathList (array)
--			- time
--			\ recentEventList
--				{ timeOffset, health, healthChange, maximumHealth, message, type }

-----------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("AssessmentDeaths")
L:RegisterTranslations("enUS", function() return {
	["Deaths List"] = true,
	["Deaths"] = true,
	["Death Times"] = true,
	["Kill List"] = true,
	["Kill Times"] = true,

	["Death"] = true,
	["Death Options."] = true,

	["Melee"] = true,
	
	["Death Recent Event Log"] = true,
	
	["Group Only"] = true,
	["Only record kills caused by group members."] = true,

	["Death Event Log History Time"] = true,
	["Select the amount of time to buffer events to display leading up to a group member's death.\nUse 0 to disable event log recording."] = true,
} end)

L:RegisterTranslations("koKR", function() return {
	["Deaths List"] = "죽음 목록",
	["Deaths"] = "죽음",
	["Death Times"] = "죽은 횟수",
	["Kill List"] = "죽임 목록",
	["Kill Times"] = "죽임 횟수",

	["Death"] = "죽음",
	["Death Options."] = "죽음 설정입니다.",
	["Group Only"] = "파티만",
	["Only record kills caused by group members."] = "파티원에 의한 죽임만 기록합니다.",
} end)

-----------------------------------------------------------------------------

local AssessmentDeaths = Assessment:NewModule("Deaths", "AceEvent-2.0", "Parser-3.0")
AssessmentDeaths.db = Assessment:AcquireDBNamespace("Deaths")
Assessment:RegisterDefaults("Deaths", "profile", {
	enable				= true,
	gatherGroupOnly		= false,
	eventLogTime		= 10,
})

local profile
local emptyMethod = Assessment.EmptyMethod
local new, del, reset = Assessment.NewTable, Assessment.ReleaseTable, Assessment.ResetTable
local pairs = pairs
local time, GetTime = time, GetTime
local UnitHealth, UnitHealthMax = UnitHealth, UnitHealthMax

-----------------------------------------------------------------------------

Assessment.TagData.Deaths =
{
	method = function(data)
				local fightData = data.fightData
				if not fightData then return 0 end
				
				local deaths = fightData.deaths
				if not deaths then return 0 end
				
				local playerDeaths = deaths[data.key]
				if not playerDeaths then return 0 end

				return playerDeaths.count
			 end,
	data = "deaths",
	guaranteeResult = true,
	numeric = true,
}

-----------------------------------------------------------------------------

Assessment.ListType.DeathsList =
{
	method = function(fightData)
				if not fightData or not fightData.deaths then return emptyMethod end
				return pairs(fightData.deaths)
			 end,
	data = "deaths",
	description = L["Deaths List"],
}

-----------------------------------------------------------------------------

Assessment.ListType.EnumerateTable.updateData.deaths = "deaths"

-----------------------------------------------------------------------------

Assessment.DefaultDisplaySets.DeathRecentEventLog =
{
	titleTag			= "Death #[WindowParameter(2)], [WindowParameter(1)]",
	listType			= "EnumerateTable",
	listDataTag			= "deaths,[WindowParameter(1)],deathList,[WindowParameter(2)],recentEventList",
	barValueTag			= "Select(GetValueField(\"type\") == \"Damage\" || GetValueField(\"type\") == \"Heal\", GetValueField(\"health\")/GetValueField(\"maximumHealth\"), 0)", 
	normalizeBarValues	= false,
	sortKeyTag			= "Rank",
	descendingSort		= false,
	fightIndex			= -1,
	text				= { count = 5, 
							{ tag = "[GetValueField(\"timeOffset\"):Format(\"%.1f\")]s", alignment = "LEFT", offsets = { left = 0, right = 25 } },
							{ tag = "[GetValueField(\"type\") == \"Fade\":\"- \"][GetValueField(\"type\") == \"Aura\":\"+ \"][GetValueField(\"sourceName\"):Suffix(\": \")][GetValueField(\"abilityName\")]", alignment = "LEFT", offsets = { left = 25, right = -50 } },
							{ tag = "[GetValueField(\"health\"):HideIf(GetValueField(\"healthChange\")==0)]", alignment = "RIGHT", offsets = { left = -50, right = -25 }  },
							{ tag = "[(100*BarValue):HideZero:Percent]", visible = false, alignment = "RIGHT", offsets = { left = 25, right = 48 } },
							{ tag = "[GetValueField(\"details\") || Select(GetValueField(\"healthChange\") > 0, (\"+\"..GetValueField(\"healthChange\")):Green, GetValueField(\"healthChange\"):HideZero:Red)]", alignment = "RIGHT", offsets = { left = 0, right = 0 } },
						  },
	description			= L["Death Recent Event Log"],
	defaultColor		= { 0.8, 0.3, 0.3 },
	selectable			= false,
}

Assessment.DefaultDisplaySets.DeathTimes =
{
	titleTag			= "[WindowParameter(1)]",
	listType			= "EnumerateTable", 
	listDataTag			= "deaths,[WindowParameter(1)],deathList",
	barValueTag			= "0", 
	filterTag			= "1",
	fightIndex			= -1,
	text				= { count = 2, { tag= "Death [Rank]" }, { tag = "[GetValueField(\"time\"):FormatTime]" } },
	description			= L["Death Times"],
	selectable			= false,
	defaultColor		= { 0.8, 0.8, 0.8 },
	onClickType			= "SHOW_DISPLAY_SET",
	onClickDisplaySet	= L["Death Recent Event Log"],
	onClickData			= "[WindowParameter(1)],[Key]",
}

Assessment.DefaultDisplaySets.Deaths =
{
	titleTag			= "Deaths",
	listType			= "GroupMembers", 
	barValueTag			= "Deaths", 
	filterTag			= "BarValue ~= 0",
	fightIndex			= 0,
	description			= L["Deaths"],
	onClickType			= "SHOW_DISPLAY_SET",
	onClickDisplaySet	= L["Death Times"],
	onClickData			= "[Key]",
	defaultColor		= { 0.8, 0.8, 0.8 },
	colorTypeList		= { "Class", count = 1 },
}

Assessment.DefaultDisplaySets.KillTimes =
{
	titleTag			= "[WindowParameter(1)]",
	listType			= "EnumerateTable", 
	listDataTag			= "deaths,[WindowParameter(1)],deathTimes",
	barValueTag			= "0", 
	filterTag			= "1",
	fightIndex			= -1,
	text				= { count = 2, { tag= "Kill [Rank]" }, { tag = "[Value:FormatTime]" } },
	description			= L["Kill Times"],
	selectable			= false,
	defaultColor		= { 0.8, 0.3, 0.5 },
}

Assessment.DefaultDisplaySets.KillList =
{
	titleTag			= "Kill List",
	listType			= "DeathsList", 
	barValueTag			= "Deaths", 
	filterTag			= "~IsGroupMember",
	fightIndex			= 0,
	description			= L["Kill List"],
	onClickType			= "SHOW_DISPLAY_SET",
	onClickDisplaySet	= L["Kill Times"],
	onClickData			= "[Key]",
	defaultColor		= { 0.8, 0.3, 0.5 },
	colorTypeList		= { "Class", count = 1 },
}

-----------------------------------------------------------------------------

Assessment.Options.args.data.args.death =
{
	type = "group",
	name = L["Death"],
	desc = L["Death Options."],
	args =
	{
		groupOnly =
		{
			type = "toggle",
			name = L["Group Only"],
			desc = L["Only record kills caused by group members."],
			get = function() return profile.gatherGroupOnly end,
			set = function(v) profile.gatherGroupOnly = v end,
		},
		eventLogTime =
		{
			type = "range",
			name = L["Death Event Log History Time"],
			desc = L["Select the amount of time to buffer events to display leading up to a group member's death.\nUse 0 to disable event log recording."],
			min = 0,
			max = 30,
			step = 0.5,
			get = function() return profile.eventLogTime end,
			set = function(v) profile.eventLogTime = v end
		},
	}
}

-----------------------------------------------------------------------------

Assessment.DefaultTags.Deaths =
{
	tag			= "[Deaths]",
	fightIndex	= 0,
	description	= L["Deaths"],
}

-----------------------------------------------------------------------------

function AssessmentDeaths:OnInitialize()
	profile = self.db.profile
end

function AssessmentDeaths:OnEnable()
	self:RegisterParserEvent({ eventType = "Death" }, "OnParserDeathEvent")
	
	-- For event logs
	self:RegisterParserEvent({ eventType = "Damage", recipientID_not = false }, "OnParserDamageEvent")
	self:RegisterParserEvent({ eventType = "Heal", recipientID_not = false }, "OnParserHealEvent")
	self:RegisterParserEvent({ eventType = "Aura", recipientID_not = false }, "OnParserAuraEvent")
	self:RegisterParserEvent({ eventType = "Fade", recipientID_not = false }, "OnParserFadeEvent")
	self:RegisterParserEvent({ eventType = "Miss", recipientID_not = false }, "OnParserMissEvent")

	self:RegisterEvent("Assessment:StartNewFight", "OnNewFight")
end

function AssessmentDeaths:OnDisable()
end

-----------------------------------------------------------------------------

function AssessmentDeaths:OnParserDeathEvent(eventData)
	local recipientID = eventData.recipientID 
	if recipientID then
		-- It's a group member that died.. record it.
	else
		-- A non-group member..
		
		-- If the options say to only gather kills done by group members, check that.
		if profile.gatherGroupOnly then
			-- If it is group only, then check that the killer was in our group
			if not eventData.sourceID then return end
		else
			-- The option is to gather all kills. In this case, exit if there is a source name, to stop slain messages from counting twice
			if eventData.sourceName then return end
		end
	end

	local deathTime = time()
	local time = GetTime()

	for data in self:ForceIterateUpdateDataTables() do
		data.dataUpdateTimes.deaths = time
		local deaths = data.deaths
		if not deaths then
			deaths = new()
			data.deaths = deaths
		end

		local recipientName = eventData.recipientName
		local recipientDeath = deaths[recipientName]
		if recipientDeath then
			recipientDeath.count = recipientDeath.count + 1
		else
			recipientDeath = new()
			recipientDeath.count = 1
			deaths[recipientName] = recipientDeath
		end

		if recipientID then
			local deathList = recipientDeath.deathList
			if not deathList then
				deathList = new()
				
				local deathTimes = recipientDeath.deathTimes
				if deathTimes then
					-- Promote deathTimes array to deathList, and remove deathTimes
					for i=1,#deathTimes do
						local t = new()
						t.time = deathTime
						deathList[i] = t
					end
					
					recipientDeath.deathTimes = del(deathTimes)
				end
			
				recipientDeath.deathList = deathList
			end

			local record = new()
			record.time = deathTime
			record.referenceTime = time
			deathList[#deathList+1] = record
		
			if profile.eventLogTime > 0 then
				self:ScheduleEvent(self.OnGroupMemberDeath, 0.1, self, eventData.recipientName)
			end
		else
			local deathList = recipientDeath.deathList
			if deathList then
				local record = new()
				record.time = deathTime
				deathList[#deathList+1] = record
			else
				local deathTimes = recipientDeath.deathTimes
				if not deathTimes then
					deathTimes = new()
					recipientDeath.deathTimes = deathTimes
				end
				
				deathTimes[#deathTimes+1] = deathTime
			end		
		end
	end
end

-----------------------------------------------------------------------------

local eventLogHistory = { }
local tinsert = table.insert

function AssessmentDeaths:OnNewFight()
	reset(eventLogHistory, true)
end

local function ProcessLogEvent(eventData, amount, abilityName, details)
	local unitid = eventData.recipientID
	
	local recipientName = eventData.recipientName
	local unitHistory = eventLogHistory[recipientName]
	if not unitHistory then
		unitHistory = new()
		eventLogHistory[recipientName] = unitHistory
	end

	local time = GetTime()

	local health = UnitHealth(unitid)
	if health == unitHistory.health then
		health = unitHistory.predictedHealth
	else
		unitHistory.health = health
	end
	local maximumHealth = UnitHealthMax(unitid)

	local eventRecord = new()
	eventRecord.time			= time
	eventRecord.sourceName		= eventData.sourceName
	eventRecord.health			= health
	eventRecord.healthChange	= amount
	eventRecord.details			= details
	eventRecord.maximumHealth	= maximumHealth
	eventRecord.type			= eventData.eventType
	eventRecord.abilityName		= abilityName
	eventRecord.message			= eventData.message

	local predictedHealth = health + amount
	if predictedHealth < 0 then
		unitHistory.predictedHealth = 0
	elseif predictedHealth > maximumHealth then
		unitHistory.predictedHealth = maximumHealth
	else
		unitHistory.predictedHealth = predictedHealth
	end
	
	tinsert(unitHistory, 1, eventRecord)

	local eraseTime = time-profile.eventLogTime
	
	for i=#unitHistory, 1, -1 do
		local history = unitHistory[i]
		if history.time > eraseTime then
			return
		end
		unitHistory[i] = del(history)
	end
end

function AssessmentDeaths:OnParserDamageEvent(eventData)
	ProcessLogEvent(eventData, -eventData.amount, eventData.abilityName or L["Melee"])
end

function AssessmentDeaths:OnParserHealEvent(eventData)
	ProcessLogEvent(eventData, eventData.amount, eventData.abilityName)
end

function AssessmentDeaths:OnParserAuraEvent(eventData)
	ProcessLogEvent(eventData, 0, eventData.abilityName)
end

function AssessmentDeaths:OnParserFadeEvent(eventData)
	ProcessLogEvent(eventData, 0, eventData.abilityName)
end

function AssessmentDeaths:OnParserMissEvent(eventData)
	ProcessLogEvent(eventData, 0, eventData.abilityName or L["Melee"], eventData.missTypeLocal)
end

function AssessmentDeaths:OnGroupMemberDeath(recipientName, deathIndex)
	for data in self:ForceIterateUpdateDataTables() do
		local deaths = data.deaths
		if deaths then
			local recipientDeath = deaths[recipientName]
			if recipientDeath then
				local deathList = recipientDeath.deathList
				if deathList then
					deathList = deathList[#deathList]
					if deathList and not deathList.recentEventList then
						local recentEventList = new()
						local copyFrom = eventLogHistory[recipientName]
						local referenceTime = deathList.referenceTime
						
						if copyFrom and referenceTime then
							for i = #copyFrom, 1, -1 do
								local e = copyFrom[i]
								e.timeOffset = e.time - referenceTime
								recentEventList[#recentEventList+1] = e
								copyFrom[i] = nil
							end
						end
						
						deathList.recentEventList = recentEventList
					end
				end
			end
		end
	end
end

-----------------------------------------------------------------------------
